// ======================================================================
//
// MaterialData_d3d_5.cpp
// jeff grills
//
// copyright 1998 origin systems inc
//
// ======================================================================

#include "first.h"
#include "MaterialData_d3d_5.h"

#include "DirectDraw.h"
#include "Direct3d_5.h"
#include "Gl_d3d_5.h"
#include "Material.h"
#include "Palette.h"
#include "PixelIndex.h"
#include "uvstruct.h"

// ======================================================================

enum Format
{
	Format4444,
	Format1555,
	Format0555,
};

// ======================================================================

const float MATDIF   = 1.0f;
const float MATAMB   = 1.0f;
const int   MATRAMP  = 16;

// ======================================================================

MaterialData_d3d_5 *MaterialData_d3d_5::head;
MaterialData_d3d_5 *MaterialData_d3d_5::tail;
u_int               MaterialData_d3d_5::globalTagLRU;
u_int               MaterialData_d3d_5::globalTagFlushBatch;

// ======================================================================

void MaterialData_d3d_5::remove(void)
{
	while (head)
		freeSomeTextureSpace();
}

// ======================================================================

MaterialData_d3d_5::MaterialData_d3d_5(const StaticMaterial *staticMaterial)
{
	HRESULT       result;

	init();

	const Pixel8 * const rgb = staticMaterial->getRGBMap8();
	const Pixel8 * const alpha = staticMaterial->getAlphaMap();

	if (rgb == NULL && alpha == NULL)
	{
		// solid filled
		r = staticMaterial->getR();
		g = staticMaterial->getG();
		b = staticMaterial->getB();

		// -qq- ignoring alpha

		return;
	}

	const int         width  = staticMaterial->getWidth();
	const int         height = staticMaterial->getHeight();
	const RGB        *pal;
	
	if (staticMaterial->getPalette())
	  pal = staticMaterial->getPalette()->getRGB();
	else
		pal = NULL;
	
	int               i;

	translucency   = (alpha != NULL) || (staticMaterial->getA());
	transparency   = rgb && staticMaterial->getAlpha() != StaticMaterial::AlphaOff;

	Format format;

	// pick the best texture format available for this texture
	if (translucency && GL_d3d_5::supports4444())
		format = Format4444;
	else
		if (transparency && GL_d3d_5::supports1555())
			format = Format1555;
		else
			if (GL_d3d_5::supports0555())
				format = Format0555;
			else
				if (GL_d3d_5::supports1555())
					format = Format1555;
				else
					format = Format4444;

	int aSizeShift, aPlaceShift, rSizeShift, rPlaceShift, gSizeShift, gPlaceShift, bSizeShift, bPlaceShift;

	switch (format)
	{
		case Format4444:
			aSizeShift  = 4;
			aPlaceShift = 12;
			rSizeShift  = 4;
			rPlaceShift = 8;
			gSizeShift  = 4;
			gPlaceShift = 4;
			bSizeShift  = 4;
			bPlaceShift = 0;
			break;

		case Format1555:
			aSizeShift  = 7;
			aPlaceShift = 15;
			rSizeShift  = 3;
			rPlaceShift = 10;
			gSizeShift  = 3;
			gPlaceShift = 5;
			bSizeShift  = 3;
			bPlaceShift = 0;
			break;

		case Format0555:
			aSizeShift  = 8;
			aPlaceShift = 0;
			rSizeShift  = 3;
			rPlaceShift = 10;
			gSizeShift  = 3;
			gPlaceShift = 5;
			bSizeShift  = 3;
			bPlaceShift = 0;
			break;

		default:
			aSizeShift  = 8;
			aPlaceShift = 0;
			rSizeShift  = 8;
			rPlaceShift = 0;
			gSizeShift  = 8;
			gPlaceShift = 0;
			bSizeShift  = 8;
			bPlaceShift = 0;
			DEBUG_FATAL(true, ("mat_d3d_5:: bad format"));
			break;
	}

	if (!rgb && format != Format4444)
	{
		ignore = true;
		return;
	}

	// fill out the surface description
	DDSURFACEDESC sdesc;
	memset(&sdesc, 0, sizeof(sdesc));
	sdesc.dwSize = sizeof(sdesc);
	sdesc.dwFlags = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH | DDSD_PIXELFORMAT;
	sdesc.dwWidth = width;
	sdesc.dwHeight = height;
	sdesc.ddsCaps.dwCaps = DDSCAPS_SYSTEMMEMORY | DDSCAPS_TEXTURE;
	sdesc.ddpfPixelFormat.dwSize = sizeof(sdesc.ddpfPixelFormat);
	sdesc.ddpfPixelFormat.dwRGBBitCount = 16;
	sdesc.ddpfPixelFormat.dwRBitMask        = ((255) >> rSizeShift) << rPlaceShift;
	sdesc.ddpfPixelFormat.dwGBitMask        = ((255) >> gSizeShift) << gPlaceShift;
	sdesc.ddpfPixelFormat.dwBBitMask        = ((255) >> bSizeShift) << bPlaceShift;
	sdesc.ddpfPixelFormat.dwRGBAlphaBitMask = ((255) >> aSizeShift) << aPlaceShift;
	sdesc.ddpfPixelFormat.dwFlags = DDPF_RGB | (sdesc.ddpfPixelFormat.dwRGBAlphaBitMask ? DDPF_ALPHAPIXELS : 0);

	// create the texture surface
	DirectDraw::createSurface(&softSurface, &sdesc, NULL, false);

	// set a colorkey if there is transparency
	if (transparency)
	{
		if (format == Format1555 || format == Format4444)
		{
			colorKeyColor = 0; 
			// colorKeyColor = static_cast<u_short>((255 >> aSizeShift) << aPlaceShift);
		}
		else
		{
			colorKeyColor = static_cast<Pixel16>(
				((255      >> aSizeShift) << aPlaceShift) | 
				((pal[0].r >> rSizeShift) << rPlaceShift) | 
				((pal[0].g >> gSizeShift) << gPlaceShift) |
				((pal[0].b >> bSizeShift) << bPlaceShift));
		}

		// set the color key on the surface
		DDCOLORKEY colorKey;
		colorKey.dwColorSpaceLowValue = colorKey.dwColorSpaceHighValue = colorKeyColor;
		do
		{
			result = softSurface->SetColorKey(DDCKEY_SRCBLT, &colorKey);
		} while (result == DDERR_WASSTILLDRAWING);
		FATAL_HR("mat_d3d_5:: setColorKey %d", result);
	}

	// lock the surface
	memset(&sdesc, 0, sizeof(sdesc));
	sdesc.dwSize = sizeof(sdesc);
	do
	{
		result = softSurface->Lock(NULL, &sdesc, DDLOCK_WAIT, NULL);
	} while (result == DDERR_SURFACEBUSY);
	FATAL_HR("mat_d3d_5:: lock %d", result);

	// convert from the palettized source to the proper surface format
	Pixel16 *p;
	int      a, r, g, b, k, pi;
	r = staticMaterial->getR() >> rSizeShift;
	g = staticMaterial->getG() >> gSizeShift;
	b = staticMaterial->getB() >> bSizeShift;
	a = staticMaterial->getA();
	for (i = 0, k = 0; i < height; ++i)
	{
		p = pixelIndex(reinterpret_cast<Pixel16 *>(sdesc.lpSurface), 0, i, sdesc.lPitch);

		for (int j = 0; j < width; ++j, ++p, ++k)
		{
			if (transparency && rgb[k] == 0)
				*p = colorKeyColor;
			else
			{
				if (format == Format4444 && alpha)
					a = alpha[k];
				
				if (rgb)
				{
					pi = rgb[k];
					r = pal[pi].r >> rSizeShift;
					g = pal[pi].g >> gSizeShift;
					b = pal[pi].b >> bSizeShift;
				}

				*p = static_cast<Pixel16>(
					((a >> aSizeShift) << aPlaceShift) | 
					(r << rPlaceShift) | 
					(g << gPlaceShift) | 
					(b << bPlaceShift));
			}
		}	
	}

	// unlock the texture
	result = softSurface->Unlock(sdesc.lpSurface);
	FATAL_HR("mat_d3d_5:: unlock %d", result);

	// get the texture interface from the surface
	result = softSurface->QueryInterface(IID_IDirect3DTexture2, (LPVOID *)&softTexture);
	FATAL_HR("mat_d3d_5:: queryInt %d", result);
}

// ----------------------------------------------------------------------

MaterialData_d3d_5::MaterialData_d3d_5(const uv_instance *uv, u_int32 squareXOffset, u_int32 squareYOffset)
{
	HRESULT       result;

	init();
	deleteOnUnload = true;

	UNREF(uv);

	const int         width  = 256;
	const int         height = 256;
	Format            format;

	// pick the best texture format available for this texture
	if (GL_d3d_5::supports0555())
		format = Format0555;
	else
		if (GL_d3d_5::supports1555())
			format = Format1555;
		else
			format = Format4444;

	int aSizeShift, aPlaceShift, rSizeShift, rPlaceShift, gSizeShift, gPlaceShift, bSizeShift, bPlaceShift;

	switch (format)
	{
		case Format4444:
			aSizeShift  = 4;
			aPlaceShift = 12;
			rSizeShift  = 4;
			rPlaceShift = 8;
			gSizeShift  = 4;
			gPlaceShift = 4;
			bSizeShift  = 4;
			bPlaceShift = 0;
			break;

		case Format1555:
			aSizeShift  = 7;
			aPlaceShift = 15;
			rSizeShift  = 3;
			rPlaceShift = 10;
			gSizeShift  = 3;
			gPlaceShift = 5;
			bSizeShift  = 3;
			bPlaceShift = 0;
			break;

		case Format0555:
			aSizeShift  = 8;
			aPlaceShift = 0;
			rSizeShift  = 3;
			rPlaceShift = 10;
			gSizeShift  = 3;
			gPlaceShift = 5;
			bSizeShift  = 3;
			bPlaceShift = 0;
			break;

		default:
			aSizeShift  = 8;
			aPlaceShift = 0;
			rSizeShift  = 8;
			rPlaceShift = 0;
			gSizeShift  = 8;
			gPlaceShift = 0;
			bSizeShift  = 8;
			bPlaceShift = 0;
			DEBUG_FATAL(true, ("mat_d3d_5:: bad format"));
			break;
	}

	// fill out the surface description
	DDSURFACEDESC sdesc;
	memset(&sdesc, 0, sizeof(sdesc));
	sdesc.dwSize = sizeof(sdesc);
	sdesc.dwFlags = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH | DDSD_PIXELFORMAT;
	sdesc.dwWidth = width;
	sdesc.dwHeight = height;
	sdesc.ddsCaps.dwCaps = DDSCAPS_SYSTEMMEMORY | DDSCAPS_TEXTURE;
	sdesc.ddpfPixelFormat.dwSize = sizeof(sdesc.ddpfPixelFormat);
	sdesc.ddpfPixelFormat.dwRGBBitCount = 16;
	sdesc.ddpfPixelFormat.dwRBitMask        = ((255) >> rSizeShift) << rPlaceShift;
	sdesc.ddpfPixelFormat.dwGBitMask        = ((255) >> gSizeShift) << gPlaceShift;
	sdesc.ddpfPixelFormat.dwBBitMask        = ((255) >> bSizeShift) << bPlaceShift;
	sdesc.ddpfPixelFormat.dwRGBAlphaBitMask = ((255) >> aSizeShift) << aPlaceShift;
	sdesc.ddpfPixelFormat.dwFlags = DDPF_RGB | (sdesc.ddpfPixelFormat.dwRGBAlphaBitMask ? DDPF_ALPHAPIXELS : 0);

	// create the texture surface
	DirectDraw::createSurface(&softSurface, &sdesc, NULL, false);

	// lock the surface
	memset(&sdesc, 0, sizeof(sdesc));
	sdesc.dwSize = sizeof(sdesc);
	do
	{
		result = softSurface->Lock(NULL, &sdesc, DDLOCK_WAIT, NULL);
	} while (result == DDERR_SURFACEBUSY);
	FATAL_HR("mat_d3d_5:: lock %d", result);

	// copy from movie frame to here
	Pixel16 *p;
	u_int32 i;
	u_int32 quickWidth=uv->width;
	u_int32 quickHeight=uv->height;

	if (uv->options.option_flags & (X_DOUBLE | X_SKIP))  // adjust from # of pixels to actual size of the buffer
		quickWidth*=2;
	if (uv->options.option_flags & (Y_DOUBLE | Y_SKIP))
		quickHeight*=2;
	{
		u_int32 cappedX,cappedY;
		u_int32 srcIndex;

		cappedX=min(squareXOffset+256,quickWidth);  // don't extend beyond the edges of the movie source
		cappedY=min(squareYOffset+256,quickHeight);

		if (uv->options.option_flags & Y_SKIP)  // if Y is skipped
		{
			srcIndex=((quickWidth*squareYOffset)>>1)+squareXOffset;

			for (i = 0; i < cappedY-squareYOffset; i++, srcIndex+=quickWidth)
			{
				p = pixelIndex(reinterpret_cast<Pixel16 *>(sdesc.lpSurface), 0, i, sdesc.lPitch);
				memcpy(p,reinterpret_cast<Pixel16 *>(uv->last_buffer)+srcIndex,(cappedX-squareXOffset)*sizeof(Pixel16));
				i++;
				p = pixelIndex(reinterpret_cast<Pixel16 *>(sdesc.lpSurface), 0, i, sdesc.lPitch);
				memset(p,0,(cappedX-squareXOffset)*sizeof(Pixel16));
			}
		}
		else if (uv->options.option_flags & Y_DOUBLE)  // if Y is doubled, don't index further into the source, cuz it isn't there
		{
			srcIndex=((quickWidth*squareYOffset)>>1)+squareXOffset;

			for (i = 0; i < cappedY-squareYOffset; i+=2, srcIndex+=quickWidth*2)
			{
				p = pixelIndex(reinterpret_cast<Pixel16 *>(sdesc.lpSurface), 0, i, sdesc.lPitch);
				memcpy(p,reinterpret_cast<Pixel16 *>(uv->last_buffer)+srcIndex,sizeof(Pixel16)*(cappedX-squareXOffset));
				memcpy(pixelIndex(reinterpret_cast<Pixel16 *>(sdesc.lpSurface), 0, i+1, sdesc.lPitch),p,sizeof(Pixel16)*(cappedX-squareXOffset));
			}
		}
		else 
		{
			srcIndex=quickWidth*squareYOffset+squareXOffset;

			for (i = 0; i < cappedY-squareYOffset; ++i, srcIndex+=quickWidth)  // no Y doubling or skipping
			{
				p = pixelIndex(reinterpret_cast<Pixel16 *>(sdesc.lpSurface), 0, i, sdesc.lPitch);
				memcpy(p,reinterpret_cast<Pixel16 *>(uv->last_buffer)+srcIndex,sizeof(Pixel16)*(cappedX-squareXOffset));
			}
		}
	}

	// unlock the texture
	result = softSurface->Unlock(sdesc.lpSurface);
	FATAL_HR("mat_d3d_5:: unlock %d", result);

	// get the texture interface from the surface
	result = softSurface->QueryInterface(IID_IDirect3DTexture2, (LPVOID *)&softTexture);
	FATAL_HR("mat_d3d_5:: queryInt %d", result);
}

// ----------------------------------------------------------------------

MaterialData_d3d_5::~MaterialData_d3d_5(void)
{
	unload();

	if (softTexture)
	{
		softTexture->Release();
		softTexture = NULL;
	}

	if (softSurface)
	{
		softSurface->Release();
		softSurface = NULL;
	}
}

// ----------------------------------------------------------------------

void MaterialData_d3d_5::init(void)
{
	softSurface    = NULL;
	softTexture    = NULL;
	hardSurface    = NULL;
	hardTexture    = NULL;
	textureHandle  = NULL;
	colorKeyColor  = 0;
	translucency   = false;
	transparency   = false;
	ignore         = false;
	deleteOnUnload = false;
	next           = NULL;
	tagLRU         = 0;
	r              = 255;
	g              = 255;
	b              = 255;
}

// ----------------------------------------------------------------------

void MaterialData_d3d_5::freeSomeTextureSpace(void)
{
	MaterialData_d3d_5 *front, *oldest;

	DEBUG_FATAL(!head, ("mat_d3d_5::freeSomeTextureSpace no textures loaded"));
		
	if (head)
	{
		for (oldest = head, front = head->next; front; front = front->next)
			if (front->tagLRU < oldest->tagLRU)
				oldest = front;

		if (oldest->deleteOnUnload)
			delete oldest;
		else
			oldest->unload();
	}
}

// ----------------------------------------------------------------------

void MaterialData_d3d_5::load(LPDIRECT3DDEVICE2 device)
{
	HRESULT  result;

	if (GL_d3d_5::isHardware())
	{
		int                width, height;
		DDSURFACEDESC      sd;
		DDPIXELFORMAT      pixelFormat;

		// get info about the source texture
		memset(&sd, 0, sizeof(sd));
		sd.dwSize = sizeof(sd);
		result = softSurface->GetSurfaceDesc(&sd);
		FATAL_HR("mat_d3d_5::load getSurfDesc %d", result);
		width  = sd.dwWidth;
		height = sd.dwHeight;
		pixelFormat = sd.ddpfPixelFormat;

		// set the texture type
		memset(&sd, 0, sizeof(sd));
		sd.dwSize = sizeof(sd);
		sd.dwFlags = DDSD_CAPS | DDSD_PIXELFORMAT | DDSD_WIDTH | DDSD_HEIGHT;
		sd.dwWidth = width;
		sd.dwHeight = height;
		sd.ddsCaps.dwCaps = DDSCAPS_ALLOCONLOAD | DDSCAPS_VIDEOMEMORY | DDSCAPS_TEXTURE;
		sd.ddpfPixelFormat = pixelFormat;

		// create the hard surface
		while (!DirectDraw::createSurface(&hardSurface, &sd, NULL, true))
			freeSomeTextureSpace();

		// get the texture interface for hard surface
		result = hardSurface->QueryInterface(IID_IDirect3DTexture2, (LPVOID *)&hardTexture);
		FATAL_HR("mat_d3d_5::load queryInt %d", result);

		// load the system texture into the hardware
		do
		{
			result = hardTexture->Load(softTexture);
		} while (result == DDERR_SURFACEBUSY);
		FATAL_HR("mat_d3d_5::load load %d", result);
	}
	else
	{
		softTexture->AddRef();
		hardTexture = softTexture;

		softSurface->AddRef();
		hardSurface = softSurface;
	}

	// get the texture handle for the device
	result = hardTexture->GetHandle(device, &textureHandle);
	FATAL_HR("mat_d3d_5::load geth %d", result);

	// add to the linked list of loaded textures
	if (tail)
		tail->next = this;
	else
		head = this;
	tail = this;
}

// ----------------------------------------------------------------------

void MaterialData_d3d_5::unload(void)
{
	if (textureHandle)
	{
		MaterialData_d3d_5 *back, *front;

		// if we are unloading a texture that has been used since the last flush or endscene,
		// we need to flush the pipeline before we unload the texture so we do not unload
		// a texture currently in use
		if (tagLRU >= globalTagFlushBatch)
			flushBatch(false);

		// search for this material in the list of loaded materials
		for (back = NULL, front = head; front != this; back = front, front = front->next)
			;

		DEBUG_FATAL(!front, ("mat_d3d_5::unload not found"));

		// remove me from the list
		if (back)
		{
			back->next = next;
			if (tail == this)
				tail = back;
		}
		else
		{
			head = next;
			if (head == NULL)
				tail = NULL;
		}

		next = NULL;

		// free hardware texture
		hardTexture->Release();
		hardTexture   = NULL;

		// free hardware surface
		hardSurface->Release();
		hardSurface   = NULL;

		// no texture handle anymore
		textureHandle = NULL;
	}
}

// ----------------------------------------------------------------------

void MaterialData_d3d_5::flushBatch(bool endScene)
{
	globalTagFlushBatch = ++globalTagLRU;

	// flush the 3d pipeline
	if (!endScene)
	{
		HRESULT result;

		result = GL_d3d_5::getDevice()->SetRenderState(D3DRENDERSTATE_FLUSHBATCH, 0);
		FATAL_HR("mat_d3d_5::flushBatch %d\n", result);
	}
}

// ----------------------------------------------------------------------

void MaterialData_d3d_5::setup(LPDIRECT3DDEVICE2 device)
{
	HRESULT result;

	// load texture as necessary
	if (textureHandle == NULL && softTexture)
		load(device);

	// set the texture handle
	result = device->SetRenderState(D3DRENDERSTATE_TEXTUREHANDLE, textureHandle);
	FATAL_HR("mat_d3d_5::setup txhandle %d\n", result);

	// set the alpha blender
	result = device->SetRenderState(D3DRENDERSTATE_ALPHABLENDENABLE, translucency ? TRUE : FALSE);
	FATAL_HR("mat_d3d_5::setup alphaBlend\n", result);

	tagLRU = ++globalTagLRU;
}

// ======================================================================
